// Jade MIF backend
// Copyright (c) 1997 Kathleen Marszalek
// Version 1.0 (beta 2b)
// Date: October 26, 1997

#ifndef MifFOTBuilderPrivate_INCLUDED
#define MifFOTBuilderPrivate_INCLUDED 1

#include "MifDoc.h"
#include "OutputByteStream.h"
#include "FOTBuilder.h"
#include "Message.h"
#include "CmdLineApp.h"
#include "IQueue.h"
#include <assert.h>

#ifdef DSSSL_NAMESPACE
namespace DSSSL_NAMESPACE {
#endif

class MifFOTBuilder : public SerialFOTBuilder {

  public:
    MifFOTBuilder( const String<CmdLineApp::AppChar> &fileLoc,
                   const Ptr<ExtendEntityManager> &, const CharsetInfo &, CmdLineApp *app );
    ~MifFOTBuilder() {}
    static MifFOTBuilder &curInstance() { assert( CurInstance != NULL ); return *CurInstance; }
    static MifFOTBuilder *CurInstance;

    struct IndexEntryNIC {
        IndexEntryNIC()
         : pageNumber( true ), startsPageRange( false ), endsPageRange( false ) {}
        StringC sortString;
        Vector<StringC> components;
        bool pageNumber;
        bool startsPageRange;
        bool endsPageRange;
    };

    class MifExtensionFlowObj : public FOTBuilder::ExtensionFlowObj {
      public:
        virtual void atomic( MifFOTBuilder &, const NodePtr & ) const = 0;
    };

    class IndexEntryFlowObj : public MifExtensionFlowObj {
      public:
        void atomic( MifFOTBuilder &fb, const NodePtr &node ) const
         { fb.indexEntry( nic ); }
        bool hasNIC( const StringC &name ) const;
        void setNIC( const StringC &name, const Value &value );
        ExtensionFlowObj *copy() const { return new IndexEntryFlowObj(*this); }
      private:
        IndexEntryNIC nic;
    };

    void extension( const ExtensionFlowObj &fo, const NodePtr & );
    void start();
    void end();

    void startSimplePageSequence();
    void endSimplePageSequence();
    void startSimplePageSequenceHeaderFooter( unsigned );
    void endSimplePageSequenceHeaderFooter( unsigned );
    void endAllSimplePageSequenceHeaderFooter();

    void indexEntry( const IndexEntryNIC & );

    void setPageNColumns( long );
    void setPageColumnSep( Length );
    void setPageBalanceColumns( bool );

    void startNode( const NodePtr &, const StringC &, RuleType );
    void endNode();
    void currentNodePageNumber( const NodePtr & );

    void startLineField( const LineFieldNIC & );
    void endLineField();
    void doStartParagraph( const ParagraphNIC &, bool servesAsWrapper = false,
                           MifDoc::T_dimension = 0, bool allowNegativeLeading = false );
    void startParagraph( const ParagraphNIC &nic ) { doStartParagraph( nic ); }
    void doEndParagraph( bool sustainFormatStack = false, bool sustainDisplayStack = false,
                         bool paragraphBreakTest = true, bool discardThisPara = false );
    void endParagraph() { doEndParagraph(); }
    void startDisplayGroup( const DisplayGroupNIC & );
    void endDisplayGroup();
    void characters( const Char *, size_t );
    void paragraphBreak( const ParagraphNIC & );
    void externalGraphic( const ExternalGraphicNIC & );
    void rule( const RuleNIC & );
    void pageNumber();
    void startScore( Symbol );
    void endScore();
    void startLeader( const LeaderNIC & );
    void endLeader();

    void startTable(const TableNIC &);
    void endTable();
    void startTablePartSerial(const TablePartNIC &);
    void endTablePartSerial();
    void startTablePartHeader();
    void endTablePartHeader();
    void startTablePartFooter();
    void endTablePartFooter();
    void tableColumn(const TableColumnNIC &);
    void startTableRow();
    void endTableRow();
    void startTableCell(const TableCellNIC &);
    void endTableCell();

    void tableBeforeRowBorder();
    void tableAfterRowBorder();
    void tableBeforeColumnBorder();
    void tableAfterColumnBorder();
    void tableCellBeforeRowBorder();
    void tableCellAfterRowBorder();
    void tableCellBeforeColumnBorder();
    void tableCellAfterColumnBorder();
//  void setCellBeforeRowMargin(Length);
//  void setCellAfterRowMargin(Length);
//  void setCellBeforeColumnMargin(Length);
//  void setCellAfterColumnMargin(Length);
//  void setCellBackground(bool);
    void setCellRowAlignment( Symbol );
    void setFontSize(Length);
    void setFontFamilyName( const StringC & );
    void setFontWeight( Symbol );
    void setFontPosture( Symbol );
    void setStartIndent( const LengthSpec & );
    void setEndIndent( const LengthSpec & );
    void setFirstLineStartIndent( const LengthSpec & );
//  void setLastLineEndIndent(const LengthSpec &);
    void setLineSpacing( const LengthSpec & );
    void setFieldWidth( const LengthSpec & );
    void setLines( Symbol );
    void setQuadding( Symbol );
    void setDisplayAlignment( Symbol );
    void setFieldAlign( Symbol );
//  void setColor(const DeviceRGBColor &);
//  void setBackgroundColor(const DeviceRGBColor &);
//  void setBackgroundColor();
    void setPageWidth( Length );
    void setPageHeight( Length );
    void setLeftMargin( Length );
    void setRightMargin( Length );
    void setTopMargin( Length );
    void setBottomMargin( Length );
    void setHeaderMargin( Length );
    void setFooterMargin( Length );
    void setBorderPresent(bool);
    void setLineThickness(Length);
    void setCellBeforeRowMargin( Length );
    void setCellAfterRowMargin( Length );
    void setCellBeforeColumnMargin( Length );
    void setCellAfterColumnMargin( Length );
    void setLineSep(Length);
//  void setBoxSizeBefore(Length);
//  void setBoxSizeAfter(Length);
    void setPositionPointShift( const LengthSpec & );
//  void setStartMargin(const LengthSpec &);
//  void setEndMargin(const LengthSpec &);
//  void setSidelineSep(const LengthSpec &);
//  void setAsisWrapIndent(const LengthSpec &);
//  void setLineNumberSep(const LengthSpec &);
//  void setLastLineJustifyLimit(const LengthSpec &);
//  void setJustifyGlyphSpaceMaxAdd(const LengthSpec &);
//  void setJustifyGlyphSpaceMaxRemove(const LengthSpec &);
//  void setTableCornerRadius(const LengthSpec &);
//  void setBoxCornerRadius(const LengthSpec &);
//  void setMarginaliaSep(const LengthSpec &);
//  void setMinPreLineSpacing(const OptLengthSpec &);
//  void setMinPostLineSpacing(const OptLengthSpec &);
    void setMinLeading( const OptLengthSpec & );
//  void setInhibitLineBreaks(bool);
    void setHyphenate( bool );
    void setKern( bool );
//  void setLigature(bool);
//  void setScoreSpaces(bool);
//  void setFloatOutMarginalia(bool);
//  void setFloatOutSidelines(bool);
//  void setFloatOutLineNumbers(bool);
//  void setCellBackground(bool);
//  void setSpanWeak(bool);
//  void setIgnoreRecordEnd(bool);
//  void setNumberedLines(bool);
//  void setHangingPunct(bool);
//  void setBoxOpenEnd(bool);
//  void setTruncateLeader(bool);
//  void setAlignLeader(bool);
//  void setTablePartOmitMiddleHeader(bool);
//  void setTablePartOmitMiddleFooter(bool);
    void setBorderOmitAtBreak(bool);
//  void setPrincipalModeSimultaneous(bool);
//  void setMarginaliaKeepWithPrevious(bool);
//  void setGridEquidistantRows(bool);
//  void setGridEquidistantColumns(bool);
//  void setLineJoin(Symbol);
    void setLineCap( Symbol );
//  void setLineNumberSide(Symbol);
//  void setKernMode(Symbol);
    void setInputWhitespaceTreatment( Symbol );
//  void setFillingDirection(Symbol);
//  void setWritingMode(Symbol);
//  void setLastLineQuadding(Symbol);
//  void setMathDisplayMode(Symbol);
//  void setScriptPreAlign(Symbol);
//  void setScriptPostAlign(Symbol);
//  void setScriptMidSupAlign(Symbol);
//  void setScriptMidSubAlign(Symbol);
//  void setNumeratorAlign(Symbol);
//  void setDenominatorAlign(Symbol);
//  void setGridPositionCellType(Symbol);
//  void setGridColumnAlignment(Symbol);
//  void setGridRowAlignment(Symbol);
//  void setBoxType(Symbol);
//  void setGlyphAlignmentMode(Symbol);
//  void setBoxBorderAlignment(Symbol);
//  void setCellRowAlignment(Symbol);
//  void setBorderAlignment(Symbol);
//  void setSidelineSide(Symbol);
//  void setHyphenationKeep(Symbol);
//  void setPositionPreference(Symbol);
//  void setFontStructure(Symbol);
//  void setFontProportionateWidth(Symbol);
//  void setCellCrossed(Symbol);
//  void setMarginaliaSide(Symbol);
//  void setLayer(long);
//  void setBackgroundLayer(long);
    void setBorderPriority( long );
    void setLineRepeat( long );
    void setSpan( long );
//  void setMinLeaderRepeat(long);
    void setHyphenationRemainCharCount( long );
    void setHyphenationPushCharCount( long );
    void setWidowCount( long );
    void setOrphanCount( long );
//  void setExpandTabs(long);
    void setHyphenationLadderCount( long );
//  void setBackgroundTile(PublicId);
//  void setLineBreakingMethod(PublicId);
//  void setLineCompositionMethod(PublicId);
//  void setImplicitBidiMethod(PublicId);
//  void setGlyphSubstMethod(PublicId);
//  void setGlyphReorderMethod(PublicId);
//  void setHyphenationMethod(PublicId);
//  void setTableAutoWidthMethod(PublicId);
//  void setFontName(PublicId);
    void setLanguage (Letter2 );
    void setCountry( Letter2 );
//  void setEscapementSpaceBefore(const InlineSpace &);
//  void setEscapementSpaceAfter(const InlineSpace &);
//  void setGlyphSubstTable(const Vector<ConstPtr<GlyphSubstTable> > &);

    struct DisplayInfo : public Link {

        DisplayInfo( const DisplayNIC &nic, DisplayInfo *parentDs )
         : spaceBefore( nic.spaceBefore), spaceAfter( nic.spaceAfter ), keep( nic.keep ),
           breakAfter( nic.breakAfter ), breakBefore( nic.breakBefore ),
           keepWithPrevious( nic.keepWithPrevious ), keepWithNext( nic.keepWithNext ),
           mayViolateKeepBefore( nic.mayViolateKeepBefore ), isParagraph( false ),
           mayViolateKeepAfter( nic.mayViolateKeepAfter ), firstParaOutputed( false ),
           paragraphClosedInMif( false )
            { keepWithinPageInEffect = (nic.keep==symbolPage)
                                        ? true
                                        : (parentDs
                                            ? parentDs->keepWithinPageInEffect
                                            : false ); }
       
        DisplaySpace spaceBefore;
        DisplaySpace spaceAfter;
        Symbol keep;
        Symbol breakBefore;
        Symbol breakAfter;
        bool keepWithPrevious;
        bool keepWithNext;
        bool mayViolateKeepBefore;
        bool mayViolateKeepAfter;
        bool firstParaOutputed;
        bool isParagraph;
        bool paragraphClosedInMif;

        bool keepWithinPageInEffect;
    };

    struct DisplaySpaceInfo : public Link {

        DisplaySpaceInfo( DisplaySpace space_, Symbol breakType_, bool breakIsAfter_ )
         : space( space_ ), breakType( breakType_ ), breakIsAfter( breakIsAfter_ ) {}

        DisplaySpace space;
        Symbol breakType;
        bool breakIsAfter;
    };

    struct EffectiveDisplaySpace {

        long nominal;
        long min;
        long max;
        long priority;
        bool conditional;
        bool force;

        EffectiveDisplaySpace() { set(); }
        EffectiveDisplaySpace( long) { set(); }

        void set( long nominal_ = 0, long min_ = 0, long max_ = 0, long priority_ = 0,
                  bool conditional_ = true, bool force_ = false )
              { nominal = nominal_; min = min_; max = max_; priority = priority_;
                conditional = conditional_; force = force_; }
        void combine( const EffectiveDisplaySpace );
        void clear() { set(); }
    };

    struct TFotSimplePageSequence {

         TFotSimplePageSequence() : BodyTextFlow( NULL ), FirstHeaderTextFlow( NULL ),
                                    FirstFooterTextFlow( NULL ), LeftHeaderTextFlow( NULL ),
                                    LeftFooterTextFlow( NULL ), RightHeaderTextFlow( NULL ),
                                    RightFooterTextFlow( NULL ) {}
        MifDoc::TextFlow *BodyTextFlow;
        MifDoc::TextFlow *FirstHeaderTextFlow;
        MifDoc::TextFlow *FirstFooterTextFlow;
        MifDoc::TextFlow *LeftHeaderTextFlow;
        MifDoc::TextFlow *LeftFooterTextFlow;
        MifDoc::TextFlow *RightHeaderTextFlow;
        MifDoc::TextFlow *RightFooterTextFlow;

        MifDoc::ParagraphFormat paragraphFormat;

    };

    struct Border {

        Border( bool cellBorder_ = true )
         : cellBorder( cellBorder_ ), borderPresent( false ) {}
        long   borderPriority;
        Length lineThickness;
        bool   borderPresent;
        long   lineRepeat;
        Length lineSep;
        bool cellBorder;

        MifDoc::T_tagstring makeMifRuling( MifDoc &mifDoc );
        void resolve( Border &adjacentBorder );
        void setFromFot();
    };

    struct Column {

        Column() : hasWidth( 0 ) {}
        bool hasWidth;
        TableLengthSpec width;
    };

    struct Cell {

        Cell() : missing( false ), MifCell( new MifDoc::Cell ), OverlappingCell( NULL ),
                 nRowsSpanned( 1 ), nColumnsSpanned( 1 ) {}
        bool missing;

        unsigned nColumnsSpanned;
        unsigned nRowsSpanned;
        Border beforeRowBorder;
        Border afterRowBorder;
        Border beforeColumnBorder;
        Border afterColumnBorder;

        Cell *OverlappingCell;
        MifDoc::Cell &mifCell()
         { assert( MifCell != NULL ); return *MifCell; }
        void translate( MifDoc::Cell &mifCell, MifDoc &mifDoc );

      protected:
        MifDoc::Cell *MifCell;    
    };

    struct Row {

        Row() {}
        Vector<Cell> Cells;

        void translate( Vector<MifDoc::Row> &mifRows, MifDoc &mifDoc );
    };

    struct Table;
    struct TablePart {

        TablePart()
         : MifTableNum( 0 ), ParentTable( NULL ), columnsProcessed( false ),
           needsColumnReprocessing( false ) {}

        Vector<Column> Columns;
        Vector<Row> Header;
        Vector<Row> Body;
        Vector<Row> Footer;                

        void translate( MifDoc &mifDoc );
        void processColumns();
        void begin( Table &parentTable, MifDoc &mifDoc );

        MifDoc::Tbl &mifTable( MifDoc &mifDoc )
         { assert( MifTableNum > 0 ); return mifDoc.tbls()[MifTableNum-1]; }
        Table &parentTable() { assert( ParentTable != NULL ); return *ParentTable; }

        size_t MifTableNum;
        Table *ParentTable;
        bool columnsProcessed;
        bool needsColumnReprocessing;
    };

    struct Table {

        Table() : beforeRowBorder( false ), afterRowBorder( false ), beforeColumnBorder( false ),
                  afterColumnBorder( false ), CurCell( NULL ), CurTablePart( NULL ),
                  DefaultTblFormatGenerated( false ), NoTablePartsSeen( true ) {}

        Vector<TablePart> TableParts;

        Border beforeRowBorder;
        Border afterRowBorder;
        Border beforeColumnBorder;
        Border afterColumnBorder;
//        TableNIC nic;
        Length tableWidth;
        Symbol displayAlignment;
        Length startIndent;

        void resolveBorders( Vector<Row> &rows, bool hasFirstTableRow, bool hasLastTableRow );
        void begin( MifDoc &mifDoc );

        Vector<Row> &curRows() { assert( CurRows != NULL ); return *CurRows; }
        TablePart &curTablePart() { assert( CurTablePart != NULL ); return *CurTablePart; }
        Cell &curCell() { assert( CurCell != NULL ); return *CurCell; }

        TablePart *CurTablePart;
        Cell *CurCell;
        Vector<Row> *CurRows;
        
        bool DefaultTblFormatGenerated;
        bool NoTablePartsSeen;

    };

    struct Format : public MifDoc::ParagraphFormat {

        Format() : MifDoc::ParagraphFormat(), FotLineSpacingSpec( 12000 ),
                   FotFirstLineStartIndentSpec( 0 ), FotStartIndentSpec( 0 ),
                   FotEndIndentSpec( 0 ), FotLanguage( 0 ), FotCurDisplaySize( 0 ),
                   FotLineThickness( 1000 ), FotBorderPresent( true ), FotLineRepeat( 1 ),
                   FotLineSep( 1000 ), FotBorderPriority( 0 ),
                   FotDisplayAlignment( symbolStart ), FotLeftMargin( 1 ), FotRightMargin( 1 ),
                   FotTopMargin( 0 ), FotBottomMargin( 0 ), FotHeaderMargin( 0 ),
                   FotFooterMargin( 0 ), FotPageWidth( 72000*8 ),
                   FotPageHeight( (72000*23)/2 ), FotLineCap( symbolButt ),
                   FotPositionPointShiftSpec( 0 ), FotMinLeading(),
                   FotFieldAlign( symbolStart ), FotFieldWidth( 0 ), FotLines( symbolWrap ),
                   FotInputWhitespaceTreatment( symbolPreserve ), FotPageNColumns( 1 ),
                   FotPageColumnSep( 72000/2 ), FotSpan( 1 ), FotPageBalanceColumns( false )  {}

        LengthSpec    FotFirstLineStartIndentSpec;
        LengthSpec    FotStartIndentSpec;
        LengthSpec    FotEndIndentSpec;
        LengthSpec    FotLineSpacingSpec;
        LengthSpec    FotPositionPointShiftSpec;
        LengthSpec    FotFieldWidth;
        OptLengthSpec FotMinLeading;
        long          FotCurDisplaySize;
        Length        FotLineThickness;
        Symbol        FotLineCap;
        long          FotBorderPriority;
        bool          FotBorderPresent;
        long          FotLineRepeat;
        Length        FotLineSep;
        Symbol        FotDisplayAlignment;
        Length        FotLeftMargin;
        Length        FotRightMargin;
        Length        FotTopMargin;
        Length        FotBottomMargin;
        Length        FotHeaderMargin;
        Length        FotFooterMargin;
        Length        FotPageWidth;
        Length        FotPageHeight;
        Symbol        FotFieldAlign;
        Symbol        FotInputWhitespaceTreatment;
        Symbol        FotLines;
        long          FotPageNColumns;
        Length        FotPageColumnSep;
        long          FotSpan;
        bool          FotPageBalanceColumns;

        Letter2 FotLanguage;
        Letter2 FotCountry;
        void computePgfLanguage();

        static const Length INITIAL_PAGE_SIZE() { return 72000*8; }
    };

    struct NodeInfo {

      NodeInfo() {}
      NodeInfo( NodePtr node_, unsigned nodeLevel_ )
       : node( node_ ), nodeLevel( nodeLevel_ ) {}
      NodePtr node;
      unsigned nodeLevel;
      static unsigned nonEmptyElementsOpened;
      static unsigned curNodeLevel;
    };

    void synchronizeFontFormat();
    long computeLengthSpec( const LengthSpec &spec ) const;
    Format &format() const
        { Format *result = formatStack.head(); assert( result != NULL ); return *result; }
    Table &curTable() { return CurTable; }
    void setCurLeaderStream( MifTmpOutputByteStream *s, bool doDelete = true )
     { if( doDelete && CurLeaderStream ) delete CurLeaderStream; CurLeaderStream = s; }
    MifTmpOutputByteStream &curLeaderStream()
     { assert( CurLeaderStream != NULL ); return *CurLeaderStream; }

    CmdLineApp *App;
    const CharsetInfo *SystemCharset;
    Ptr<ExtendEntityManager> EntityManager;
    bool paragraphBreakInEffect;
    bool inLeader;
    bool lastFlowObjectWasWhitespace;
    bool firstHeaderFooter;
    bool inSimplePageSequence;
    bool bookComponentOpened;
    bool bookComponentAvailable;

  protected:

    TFotSimplePageSequence FotSimplePageSequence;
    Table CurTable;

    MifDoc::T_pathname MifFOTBuilder::systemIdToMifPathname( const StringC &systemId );
    bool MifFOTBuilder::systemIdToMifPathname( const StringC &systemId, MifDoc::T_pathname & );
    int  systemIdFilename( const StringC &systemId, StringC &filename );
    void initMifBookComponent();
    void setupSimplePageSequence();
    void beginHeaderFooter( bool header );
    void beginHeader() { beginHeaderFooter( true ); }
    void beginFooter() { beginHeaderFooter( false ); };
    void endHeaderFooter();
    void makeEmptyTextFlow( MifDoc::TextRect & );
    void checkForParagraphReopening();
    void outPotentialMarkers();
    void outString( const Char *s, size_t n, MifTmpOutputByteStream *o = NULL,
                    bool inParagraph = true, String<char> *targetString = NULL );

    MifDoc::Frame &makeAnchoredFrame
     ( MifDoc::T_keyword frameType, long width, long height,
       MifDoc::T_keyword anchorAlign = MifDoc::sLeft );
    void setupHeaderFooterParagraphFormat( MifDoc::ParagraphFormat &,
                                           MifDoc::ParagraphFormat &,
                                           MifDoc::T_dimension );
    OutputByteStream &os() { return *outputStream; }

    MifDoc mifDoc;
    OutputByteStream *outputStream;
    MifTmpOutputByteStream *CurLeaderStream;
    EffectiveDisplaySpace createEffectiveDisplaySpace( const DisplaySpace &ds ) const;
    IList<DisplayInfo> displayStack;
    IList<Format> formatStack;
    IQueue<DisplaySpaceInfo> displaySpaceQueue;
    Vector<NodeInfo> nodeStack;

    Format nextFormat;

    Symbol pendingBreak;
    EffectiveDisplaySpace pendingEffectiveDisplaySpace;
    void processDisplaySpaceStack();

    void startDisplay( const DisplayNIC & );
    void endDisplay();

    // sytem font code from James Clark
    enum { CHAR_TABLE_CHAR_BITS = 16 };
    enum { CHAR_TABLE_SYMBOL_FLAG = 1U << 31 };
    CharMap<Unsigned32> CharTable;

    struct SymbolFont {
        const char *name;
        Char mapping[256];
    };

    enum { nSymbolFonts = 3 };
    static const SymbolFont SymbolFonts[nSymbolFonts];
};

#ifdef DSSSL_NAMESPACE
}
#endif

#endif /* not MifFOTBuilderPrivate_INCLUDED */
